@@ -81,7 +81,7 @@ extern int cmsDelayDefault;
 /* ---------------------------- included code files ------------------------ */
 
 /* ---------------------------- local types -------------------------------- */
-
+typedef enum {FakeMouseEvent, FakeKeyEvent} FakeEventType;
 /* ---------------------------- forward declarations ----------------------- */
 
 /* ---------------------------- local variables ---------------------------- */
@@ -800,7 +800,7 @@ static void SetLayerButtonFlag(
 		{
 			add = 1;
 		}
-		for (i = start; i < NUMBER_OF_BUTTONS; i += add)
+		for (i = start; i < NUMBER_OF_TITLE_BUTTONS; i += add)
 		{
 			if (set)
 			{
@@ -851,7 +851,7 @@ static void SetMWMButtonFlag(
 		{
 			add = 1;
 		}
-		for (i = start; i < NUMBER_OF_BUTTONS; i += add)
+		for (i = start; i < NUMBER_OF_TITLE_BUTTONS; i += add)
 		{
 			if (set)
 			{
@@ -900,7 +900,7 @@ static void do_button_style(F_CMD_ARGS, Bool do_add)
 		button = BUTTON_INDEX(button);
 	}
 
-	if (parm == NULL || button >= NUMBER_OF_BUTTONS || button < 0)
+	if (parm == NULL || button >= NUMBER_OF_TITLE_BUTTONS || button < 0)
 	{
 		fvwm_msg(
 			ERR, "ButtonStyle", "Bad button style (1) in line %s",
@@ -954,7 +954,7 @@ static void do_button_style(F_CMD_ARGS, Bool do_add)
 	}
 	else
 	{
-		for (i = 0; i < NUMBER_OF_BUTTONS; ++i)
+		for (i = 0; i < NUMBER_OF_TITLE_BUTTONS; ++i)
 		{
 			if (((multi & 1) && !(i & 1)) ||
 			    ((multi & 2) && (i & 1)))
@@ -987,7 +987,7 @@ static void do_button_style(F_CMD_ARGS, Bool do_add)
 					if (multi)
 					{
 						for (i = 0;
-						     i < NUMBER_OF_BUTTONS; ++i)
+						     i < NUMBER_OF_TITLE_BUTTONS; ++i)
 						{
 							if (((multi & 1) &&
 							     !(i & 1)) ||
@@ -1098,7 +1098,7 @@ static void do_button_style(F_CMD_ARGS, Bool do_add)
 		{
 			if (multi)
 			{
-				for (i = 0; i < NUMBER_OF_BUTTONS; ++i)
+				for (i = 0; i < NUMBER_OF_TITLE_BUTTONS; ++i)
 				{
 					if (((multi & 1) && !(i & 1)) ||
 					    ((multi & 2) && (i & 1)))
@@ -1180,7 +1180,7 @@ void update_decors_colorset(int cset)
 #endif
 	{
 		has_changed = 0;
-		for(i = 0; i < NUMBER_OF_BUTTONS; i++)
+		for(i = 0; i < NUMBER_OF_TITLE_BUTTONS; i++)
 		{
 			decor->flags.has_changed |= update_titlebutton_colorset(
 				&(decor->buttons[i]), cset);
@@ -1561,7 +1561,7 @@ Bool ReadDecorFace(char *s, DecorFace *df, int button, int verbose)
 				b = BUTTON_INDEX(b);
 				s += offset;
 			}
-			if (b >= 0 && b < NUMBER_OF_BUTTONS)
+			if (b >= 0 && b < NUMBER_OF_TITLE_BUTTONS)
 			{
 				LoadDefaultButton(df, b);
 			}
@@ -2090,7 +2090,7 @@ void InitFvwmDecor(FvwmDecor *decor)
 
 	/* initialize title-bar button styles */
 	DFS_FACE_TYPE(tmpdf.style) = SimpleButton;
-	for (i = 0; i < NUMBER_OF_BUTTONS; ++i)
+	for (i = 0; i < NUMBER_OF_TITLE_BUTTONS; ++i)
 	{
 		int j = 0;
 		for (; j < BS_MaxButtonState; ++j)
@@ -2525,7 +2525,7 @@ void CMD_Wait(F_CMD_ARGS)
 				t = NULL;
 			}
 
-			if (e.type == MapNotify)
+			if (e.type == MapNotify && e.xmap.event == Scr.Root)
 			{
 				if (!*wait_string)
 				{
@@ -2553,12 +2553,14 @@ void CMD_Wait(F_CMD_ARGS)
 			}
 			else if (e.type == KeyPress)
 			{
+				/* TODO: should I be using <t> or <exc->w.fw>?
+				 */
+				int context = GetContext(&t, t, &e, &nonewin);
 				escape = CheckBinding(
 					Scr.AllBindings, STROKE_ARG(0)
 					e.xkey.keycode, e.xkey.state,
-					GetUnusedModifiers(),
-					GetContext(&t, t, &e, &nonewin),
-					BIND_KEYPRESS);
+					GetUnusedModifiers(), context,
+					BIND_KEYPRESS, &t->class, t->name.name);
 				if (escape != NULL)
 				{
 					if (!strcasecmp(escape,"escapefunc"))
@@ -3159,7 +3161,7 @@ void CMD_DestroyDecor(F_CMD_ARGS)
 			found->flags.has_changed = 1;
 			found->flags.has_title_height_changed = 0;
 			found->titlebar.flags.has_changed = 1;
-			for (i = 0; i < NUMBER_OF_BUTTONS; ++i)
+			for (i = 0; i < NUMBER_OF_TITLE_BUTTONS; ++i)
 			{
 				TB_FLAGS(found->buttons[i]).has_changed = 1;
 			}
@@ -3328,9 +3330,7 @@ void CMD_SetEnv(F_CMD_ARGS)
 	action = GetNextToken(action, &szValue);
 	if (!szValue)
 	{
-		/* no value, ignore */
-		free(szVar);
-		return;
+		szValue = safestrdup("");
 	}
 	szPutenv = safemalloc(strlen(szVar) + strlen(szValue) + 2);
 	sprintf(szPutenv,"%s=%s", szVar, szValue);
@@ -3345,18 +3345,13 @@ void CMD_SetEnv(F_CMD_ARGS)
 void CMD_UnsetEnv(F_CMD_ARGS)
 {
 	char *szVar = NULL;
-	char *szPutenv = NULL;
 
-	action = GetNextToken(action, &szVar);
+	szVar = PeekToken(action, &action);
 	if (!szVar)
 	{
 		return;
 	}
-	szPutenv = (char *)safemalloc(strlen(szVar) + 2);
-	sprintf(szPutenv, "%s=", szVar);
-	flib_putenv(szVar, szPutenv);
-	free(szVar);
-	free(szPutenv);
+	flib_unsetenv(szVar);
 
 	return;
 }
@@ -3507,14 +3502,14 @@ void CMD_BugOpts(F_CMD_ARGS)
 			switch (toggle)
 			{
 			case -1:
-				Scr.bo.DisableConfigureNotify ^= 1;
+				Scr.bo.do_disable_configure_notify ^= 1;
 				break;
 			case 0:
 			case 1:
-				Scr.bo.DisableConfigureNotify = toggle;
+				Scr.bo.do_disable_configure_notify = toggle;
 				break;
 			default:
-				Scr.bo.DisableConfigureNotify = 0;
+				Scr.bo.do_disable_configure_notify = 0;
 				break;
 			}
 		}
@@ -3523,14 +3518,14 @@ void CMD_BugOpts(F_CMD_ARGS)
 			switch (toggle)
 			{
 			case -1:
-				Scr.bo.InstallRootCmap ^= 1;
+				Scr.bo.do_install_root_cmap ^= 1;
 				break;
 			case 0:
 			case 1:
-				Scr.bo.InstallRootCmap = toggle;
+				Scr.bo.do_install_root_cmap = toggle;
 				break;
 			default:
-				Scr.bo.InstallRootCmap = 0;
+				Scr.bo.do_install_root_cmap = 0;
 				break;
 			}
 		}
@@ -3539,17 +3534,17 @@ void CMD_BugOpts(F_CMD_ARGS)
 			switch (toggle)
 			{
 			case -1:
-				Scr.bo.ModalityIsEvil ^= 1;
+				Scr.bo.is_modality_evil ^= 1;
 				break;
 			case 0:
 			case 1:
-				Scr.bo.ModalityIsEvil = toggle;
+				Scr.bo.is_modality_evil = toggle;
 				break;
 			default:
-				Scr.bo.ModalityIsEvil = 0;
+				Scr.bo.is_modality_evil = 0;
 				break;
 			}
-			if (Scr.bo.ModalityIsEvil)
+			if (Scr.bo.is_modality_evil)
 			{
 				SetMWM_INFO(Scr.NoFocusWin);
 			}
@@ -3559,14 +3554,14 @@ void CMD_BugOpts(F_CMD_ARGS)
 			switch (toggle)
 			{
 			case -1:
-				Scr.bo.RaiseHackNeeded ^= 1;
+				Scr.bo.is_raise_hack_needed ^= 1;
 				break;
 			case 0:
 			case 1:
-				Scr.bo.RaiseHackNeeded = toggle;
+				Scr.bo.is_raise_hack_needed = toggle;
 				break;
 			default:
-				Scr.bo.RaiseHackNeeded = 0;
+				Scr.bo.is_raise_hack_needed = 0;
 				break;
 			}
 		}
@@ -3575,14 +3570,14 @@ void CMD_BugOpts(F_CMD_ARGS)
 			switch (toggle)
 			{
 			case -1:
-				Scr.bo.RaiseOverUnmanaged ^= 1;
+				Scr.bo.do_raise_over_unmanaged ^= 1;
 				break;
 			case 0:
 			case 1:
-				Scr.bo.RaiseOverUnmanaged = toggle;
+				Scr.bo.do_raise_over_unmanaged = toggle;
 				break;
 			default:
-				Scr.bo.RaiseOverUnmanaged = 0;
+				Scr.bo.do_raise_over_unmanaged = 0;
 				break;
 			}
 		}
@@ -3591,14 +3586,14 @@ void CMD_BugOpts(F_CMD_ARGS)
 			switch (toggle)
 			{
 			case -1:
-				Scr.bo.FlickeringQtDialogsWorkaround ^= 1;
+				Scr.bo.do_enable_flickering_qt_dialogs_workaround ^= 1;
 				break;
 			case 0:
 			case 1:
-				Scr.bo.FlickeringQtDialogsWorkaround = toggle;
+				Scr.bo.do_enable_flickering_qt_dialogs_workaround = toggle;
 				break;
 			default:
-				Scr.bo.FlickeringQtDialogsWorkaround = 0;
+				Scr.bo.do_enable_flickering_qt_dialogs_workaround = 0;
 				break;
 			}
 		}
@@ -3611,14 +3606,30 @@ void CMD_BugOpts(F_CMD_ARGS)
 			switch (toggle)
 			{
 			case -1:
-				Scr.bo.DisplayNewWindowNames ^= 1;
+				Scr.bo.do_display_new_window_names ^= 1;
 				break;
 			case 0:
 			case 1:
-				Scr.bo.DisplayNewWindowNames = toggle;
+				Scr.bo.do_display_new_window_names = toggle;
 				break;
 			default:
-				Scr.bo.DisplayNewWindowNames = 0;
+				Scr.bo.do_display_new_window_names = 0;
+				break;
+			}
+		}
+		else if (StrEquals(opt, "ExplainWindowPlacement"))
+		{
+			switch (toggle)
+			{
+			case -1:
+				Scr.bo.do_explain_window_placement ^= 1;
+				break;
+			case 0:
+			case 1:
+				Scr.bo.do_explain_window_placement = toggle;
+				break;
+			default:
+				Scr.bo.do_explain_window_placement = 0;
 				break;
 			}
 		}
@@ -3640,18 +3651,18 @@ void CMD_Emulate(F_CMD_ARGS)
 	style = PeekToken(action, NULL);
 	if (!style || StrEquals(style, "fvwm"))
 	{
-		Scr.gs.EmulateMWM = False;
-		Scr.gs.EmulateWIN = False;
+		Scr.gs.do_emulate_mwm = False;
+		Scr.gs.do_emulate_win = False;
 	}
 	else if (StrEquals(style, "mwm"))
 	{
-		Scr.gs.EmulateMWM = True;
-		Scr.gs.EmulateWIN = False;
+		Scr.gs.do_emulate_mwm = True;
+		Scr.gs.do_emulate_win = False;
 	}
 	else if (StrEquals(style, "win"))
 	{
-		Scr.gs.EmulateMWM = False;
-		Scr.gs.EmulateWIN = True;
+		Scr.gs.do_emulate_mwm = False;
+		Scr.gs.do_emulate_win = True;
 	}
 	else
 	{
@@ -3718,7 +3729,39 @@ void CMD_SetAnimation(F_CMD_ARGS)
 	return;
 }
 
-void CMD_FakeClick(F_CMD_ARGS)
+/* Determine which modifiers are required with a keycode to make <keysym>. */
+static Bool FKeysymToKeycode (Display *dpy, KeySym keysym,
+	unsigned int *keycode, unsigned int *modifiers)
+{
+	int m;
+
+	*keycode = XKeysymToKeycode(dpy, keysym);
+	*modifiers = 0;
+
+	for (m = 0; m <= 8; ++m)
+	{
+		KeySym ks = XKeycodeToKeysym(dpy, *keycode, m);
+		if (ks == keysym)
+		{
+			switch (m)
+			{
+				case 0: /* No modifiers */
+					break;
+				case 1: /* Shift modifier */
+					*modifiers |= ShiftMask;
+					break;
+				default:
+					fvwm_msg(ERR, "FKeysymToKeycode",
+						"Unhandled modifier %d", m);
+					break;
+			}
+			return True;
+		}
+	}
+	return False;
+}
+
+static void __fake_event(F_CMD_ARGS, FakeEventType type)
 {
 	char *token;
 	char *optlist[] = {
@@ -3732,19 +3775,19 @@ void CMD_FakeClick(F_CMD_ARGS)
 	unsigned int mask = 0;
 	Window root = Scr.Root;
 	int maxdepth = 0;
+	static char args[128];
+	strncpy(args, action, sizeof(args) - 1);
+
+	/* get the mask of pressed/released buttons/keys */
+	FQueryPointer(
+		dpy, Scr.Root, &root, &JunkRoot, &JunkX, &JunkY, &JunkX,
+		&JunkY, &mask);
 
-	/* get the mask of pressed/released buttons */
-	if (FQueryPointer(
-		    dpy, Scr.Root, &root, &JunkRoot, &JunkX, &JunkY, &JunkX,
-		    &JunkY, &mask) == False)
-	{
-		/* pointer is on a different screen - that's okay here */
-	}
 	token = PeekToken(action, &action);
 	while (token && action)
 	{
 		int index = GetTokenIndex(token, optlist, 0, NULL);
-		int val;
+		int val, depth;
 		XEvent e;
 		Window w;
 		Window child_w;
@@ -3754,13 +3797,9 @@ void CMD_FakeClick(F_CMD_ARGS)
 		int ry = 0;
 		Bool do_unset;
 		long add_mask = 0;
+		KeySym keysym = NoSymbol;
 
 		XFlush(dpy);
-		if (GetIntegerArguments(action, &action, &val, 1) != 1)
-		{
-			/* error */
-			return;
-		}
 		do_unset = True;
 		switch (index)
 		{
@@ -3770,38 +3809,69 @@ void CMD_FakeClick(F_CMD_ARGS)
 			/* fall through */
 		case 2:
 		case 3:
-			/* button press or release */
-			if (val >= 1 && val <= NUMBER_OF_MOUSE_BUTTONS)
+			/* key/button press or release */
+			if (type == FakeMouseEvent)
 			{
-				int depth = 1;
+				if ((GetIntegerArguments(
+					     action, &action, &val, 1) != 1) ||
+				    val < 1 ||
+				    val > NUMBER_OF_EXTENDED_MOUSE_BUTTONS)
+				{
+					fvwm_msg(
+						ERR, "__fake_event",
+						"Invalid button specifier in"
+						" \"%s\" for FakeClick.", args);
+					return; /* error */
+				}
+			}
+			else /* type == FakeKeyEvent */
+			{
+				char *key = PeekToken(action, &action);
+				if (key == NULL)
+				{
+					fvwm_msg(
+						ERR, "__fake_event",
+						"No keysym specifier in \"%s\""
+						" for FakeKeypress.", args);
+					return;
+				}
 
-				w = None;
-				child_w = root;
-				for (depth = 1; depth != maxdepth &&
-					     w != child_w && child_w != None;
-				     depth++)
+				/* Do *NOT* use FvwmStringToKeysym() as it is
+				 * case insensitive. */
+				keysym = XStringToKeysym(key);
+				if (keysym == NoSymbol)
 				{
-					w = child_w;
-					if (FQueryPointer(
-						    dpy, w, &root, &child_w,
-						    &rx, &ry, &x, &y,
-						    &JunkMask) == False)
-					{
-						/* pointer is on a different
-						 * screen - that's okay here */
-					}
+					fvwm_msg(
+						ERR, "__fake_event",
+						"Invalid keysym specifier (%s)"
+						" in \"%s\" for FakeKeypress.",
+						key, args);
+					return;
 				}
-				if (do_unset)
+			}
+
+			w = None;
+			child_w = root;
+			for (depth = 1;
+				 depth != maxdepth &&
+				     w != child_w && child_w != None;
+				 depth++)
+			{
+				w = child_w;
+				if (FQueryPointer(
+						dpy, w, &root, &child_w,
+						&rx, &ry, &x, &y,
+						&JunkMask) == False)
 				{
-					e.type = ButtonRelease;
-					add_mask = ButtonPressMask;
-				}
-				else
-				{
-					e.type = ButtonPress;
-					add_mask = ButtonPressMask |
-						ButtonReleaseMask;
+					/* pointer is on a different
+					 * screen - that's okay here */
 				}
+			}
+
+			if (type == FakeMouseEvent)
+			{
+				e.type = (do_unset) ?
+					ButtonRelease : ButtonPress;
 				e.xbutton.display = dpy;
 				e.xbutton.window = w;
 				e.xbutton.subwindow = None;
@@ -3814,10 +3884,15 @@ void CMD_FakeClick(F_CMD_ARGS)
 				e.xbutton.button = val;
 				e.xbutton.state = mask;
 				e.xbutton.same_screen = (Scr.Root == root);
-				FSendEvent(
-					dpy, PointerWindow, True,
-					SubstructureNotifyMask | add_mask, &e);
-				XFlush(dpy);
+				/* SS: I think this mask handling code is
+				 * buggy.
+				 * The value of <mask> is overridden during a
+				 * "wait" operation. Also why are we only using
+				 * Button1Mask? What if the user has requested
+				 * a FakeClick using some other button? */
+				/* DV: Button1Mask is actually a bit.  Shifting
+				 * it by (val -1) bits to the left gives
+				 * Button2Mask, Button3Mask etc. */
 				if (do_unset)
 				{
 					mask &= ~(Button1Mask << (val - 1));
@@ -3826,37 +3901,74 @@ void CMD_FakeClick(F_CMD_ARGS)
 				{
 					mask |= (Button1Mask << (val - 1));
 				}
+				add_mask = (do_unset) ?
+					ButtonPressMask : ButtonReleaseMask;
 			}
 			else
 			{
-				/* error */
-				return;
+				/* type == FakeKeyEvent */
+				e.type = (do_unset ? KeyRelease : KeyPress);
+				e.xkey.display = dpy;
+				e.xkey.subwindow = None;
+				e.xkey.root = root;
+				e.xkey.time = fev_get_evtime();
+				e.xkey.x = x;
+				e.xkey.y = y;
+				e.xkey.x_root = rx;
+				e.xkey.y_root = ry;
+				e.xkey.same_screen = (Scr.Root == root);
+
+				w = e.xkey.window = exc->w.w;
+
+				if (FKeysymToKeycode(
+					    dpy, keysym, &(e.xkey.keycode),
+					    &(e.xkey.state)) != True)
+				{
+					fvwm_msg(DBG, "__fake_event",
+						"FKeysymToKeycode failed");
+					return;
+				}
+				e.xkey.state |= mask;
+				add_mask = (do_unset) ?
+					KeyReleaseMask : KeyPressMask;
 			}
+			FSendEvent(dpy, w, True,
+				SubstructureNotifyMask | add_mask, &e);
+			XFlush(dpy);
 			break;
 		case 4:
 		case 5:
 			/* wait */
-			if (val > 0 && val <= 1000000)
+			if ((GetIntegerArguments(
+				     action, &action, &val, 1) != 1) ||
+			    val <= 0 || val > 1000000)
 			{
-				usleep(1000 * val);
-				if (FQueryPointer(
-					    dpy, Scr.Root, &root, &JunkRoot,
-					    &JunkX, &JunkY, &JunkX, &JunkY,
-					    &mask) == False)
-				{
-					/* pointer is on a different screen -
-					 * that's okay here */
-				}
-			}
-			else
-			{
-				/* error */
+				fvwm_msg(ERR, "__fake_event",
+					"Invalid wait value in \"%s\"", args);
 				return;
 			}
+
+			usleep(1000 * val);
+			if (FQueryPointer(
+					dpy, Scr.Root, &root, &JunkRoot,
+					&JunkX, &JunkY, &JunkX, &JunkY,
+					&mask) == False)
+			{
+				/* pointer is on a different screen -
+				 * that's okay here */
+			}
 			break;
 		case 6:
 		case 7:
 			/* set modifier */
+			if (GetIntegerArguments(action, &action, &val, 1) != 1)
+			{
+				fvwm_msg(
+					ERR, "__fake_event",
+					"Invalid modifier value in \"%s\"",
+					args);
+				return;
+			}
 			do_unset = False;
 			if (val < 0)
 			{
@@ -3884,6 +3996,8 @@ void CMD_FakeClick(F_CMD_ARGS)
 				/* error */
 				return;
 			}
+			/* SS: Could be buggy if a "modifier" operation
+			 * preceeds a "wait" operation. */
 			if (do_unset)
 			{
 				mask &= ~val;
@@ -3896,10 +4010,17 @@ void CMD_FakeClick(F_CMD_ARGS)
 		case 8:
 		case 9:
 			/* new max depth */
+			if (GetIntegerArguments(action, &action, &val, 1) != 1)
+			{
+				fvwm_msg(ERR, "__fake_event",
+					"Invalid depth value in \"%s\"", args);
+				return;
+			}
 			maxdepth = val;
 			break;
 		default:
-			/* error */
+			fvwm_msg(ERR, "__fake_event",
+				"Invalid command (%s) in \"%s\"", token, args);
 			return;
 		}
 		if (action)
@@ -3911,6 +4032,20 @@ void CMD_FakeClick(F_CMD_ARGS)
 	return;
 }
 
+void CMD_FakeClick(F_CMD_ARGS)
+{
+	__fake_event(F_PASS_ARGS, FakeMouseEvent);
+
+	return;
+}
+
+void CMD_FakeKeypress(F_CMD_ARGS)
+{
+	__fake_event(F_PASS_ARGS, FakeKeyEvent);
+
+	return;
+}
+
 /* A function to handle stroke (olicha Nov 11, 1999) */
 #ifdef HAVE_STROKE
 void CMD_StrokeFunc(F_CMD_ARGS)
@@ -3920,7 +4055,7 @@ void CMD_StrokeFunc(F_CMD_ARGS)
 	int modifiers = exc->x.etrigger->xbutton.state;
 	int start_event_type = exc->x.etrigger->type;
 	char sequence[STROKE_MAX_SEQUENCE + 1];
-	char *stroke_action;
+	char *stroke_action, *name;
 	char *opt = NULL;
 	Bool finish_on_release = True;
 	KeySym keysym;
@@ -3939,6 +4074,9 @@ void CMD_StrokeFunc(F_CMD_ARGS)
 	Bool feed_back = False;
 	int stroke_width = 1;
 	XEvent e;
+	XClassHint tmp, *pClass;
+
+
 
 	if (!GrabEm(CRS_STROKE, GRAB_NORMAL))
 	{
@@ -4206,10 +4344,21 @@ void CMD_StrokeFunc(F_CMD_ARGS)
 		return;
 	}
 
+	if (exc->w.fw == NULL)
+	{
+		tmp.res_class = tmp.res_name = name = "root";
+		pClass = &tmp;
+	}
+	else
+	{
+		pClass = &exc->w.fw->class;
+		name = exc->w.fw->name.name;
+	}
+
 	/* check for a binding */
 	stroke_action = CheckBinding(
 		Scr.AllBindings, sequence, 0, modifiers, GetUnusedModifiers(),
-		exc->w.wcontext, BIND_STROKE);
+		exc->w.wcontext, BIND_STROKE, pClass, name);
 
 	/* execute the action */
 	if (stroke_action != NULL)
