@@ -14,7 +14,7 @@
  * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
  */
 
-/* ---------------------------- included header files ----------------------- */
+/* ---------------------------- included header files ---------------------- */
 
 #include "config.h"
 
@@ -23,6 +23,8 @@
 
 #include "libs/fvwmlib.h"
 #include "libs/FShape.h"
+#include "libs/charmap.h"
+#include "libs/wcontext.h"
 #include "fvwm.h"
 #include "externs.h"
 #include "execcontext.h"
@@ -34,17 +36,18 @@
 #include "focus.h"
 #include "borders.h"
 #include "frame.h"
-extern void GetWindowSizeHints(FvwmWindow *tmp);
+#include "gnome.h"
+#include "ewmh.h"
 
-/* ---------------------------- local definitions --------------------------- */
+/* ---------------------------- local definitions -------------------------- */
 
-/* ---------------------------- local macros -------------------------------- */
+/* ---------------------------- local macros ------------------------------- */
 
-/* ---------------------------- imports ------------------------------------- */
+/* ---------------------------- imports ------------------------------------ */
 
-/* ---------------------------- included code files ------------------------- */
+/* ---------------------------- included code files ------------------------ */
 
-/* ---------------------------- local types --------------------------------- */
+/* ---------------------------- local types -------------------------------- */
 
 typedef struct
 {
@@ -73,7 +76,7 @@ typedef struct
 	Window w_with_focus;
 	int current_step;
 	int curr_titlebar_compression;
-	direction_type shade_dir;
+	direction_t shade_dir;
 	window_parts trans_parts;
 	struct
 	{
@@ -87,6 +90,7 @@ typedef struct
 		unsigned is_lazy_shading : 1;
 		unsigned is_setup : 1;
 		unsigned is_shading : 1;
+		unsigned was_moved : 1;
 	} flags;
 	/* used during the animation */
 	int next_titlebar_compression;
@@ -106,9 +110,9 @@ typedef struct
 	} step_flags;
 } mr_args_internal;
 
-/* ---------------------------- forward declarations ------------------------ */
+/* ---------------------------- forward declarations ----------------------- */
 
-/* ---------------------------- local variables ----------------------------- */
+/* ---------------------------- local variables ---------------------------- */
 
 /* windows used to hide animation steps */
 static struct
@@ -117,9 +121,9 @@ static struct
 	Window w[4];
 } hide_wins;
 
-/* ---------------------------- exported variables (globals) ---------------- */
+/* ---------------------------- exported variables (globals) --------------- */
 
-/* ---------------------------- local functions ----------------------------- */
+/* ---------------------------- local functions ---------------------------- */
 
 #if 0
 static void print_g(char *text, rectangle *g)
@@ -138,37 +142,31 @@ static void print_g(char *text, rectangle *g)
 }
 #endif
 
-static void combine_decor_gravities(
+static void combine_gravities(
 	frame_decor_gravities_type *ret_grav,
 	frame_decor_gravities_type *grav_x,
 	frame_decor_gravities_type *grav_y)
 {
-	ret_grav->decor_grav =
-		gravity_combine_xy_grav(grav_x->decor_grav,
-					grav_y->decor_grav);
-	ret_grav->title_grav =
-		gravity_combine_xy_grav(grav_x->title_grav,
-					grav_y->title_grav);
-	ret_grav->lbutton_grav =
-		gravity_combine_xy_grav(grav_x->lbutton_grav,
-					grav_y->lbutton_grav);
-	ret_grav->rbutton_grav =
-		gravity_combine_xy_grav(grav_x->rbutton_grav,
-					grav_y->rbutton_grav);
-	ret_grav->parent_grav =
-		gravity_combine_xy_grav(grav_x->parent_grav,
-					grav_y->parent_grav);
-	ret_grav->client_grav =
-		gravity_combine_xy_grav(grav_x->client_grav,
-					grav_y->client_grav);
+	ret_grav->decor_grav = gravity_combine_xy_grav(
+		grav_x->decor_grav, grav_y->decor_grav);
+	ret_grav->title_grav = gravity_combine_xy_grav(
+			grav_x->title_grav, grav_y->title_grav);
+	ret_grav->lbutton_grav = gravity_combine_xy_grav(
+		grav_x->lbutton_grav, grav_y->lbutton_grav);
+	ret_grav->rbutton_grav = gravity_combine_xy_grav(
+		grav_x->rbutton_grav, grav_y->rbutton_grav);
+	ret_grav->parent_grav = gravity_combine_xy_grav(
+		grav_x->parent_grav, grav_y->parent_grav);
+	ret_grav->client_grav = gravity_combine_xy_grav(
+		grav_x->client_grav, grav_y->client_grav);
 
 	return;
 }
 
 static void get_resize_decor_gravities_one_axis(
-	frame_decor_gravities_type *ret_grav, direction_type title_dir,
-	frame_move_resize_mode axis_mode, direction_type neg_dir,
-	direction_type pos_dir, int is_moving)
+	frame_decor_gravities_type *ret_grav, direction_t title_dir,
+	frame_move_resize_mode axis_mode, direction_t neg_dir,
+	direction_t pos_dir, int is_moving)
 {
 
 	int title_grav;
@@ -313,7 +311,7 @@ static void frame_setup_titlebar(
 	FvwmWindow *fw, rectangle *frame_g, window_parts setup_parts,
 	rectangle *diff_g)
 {
-	frame_title_layout_type title_layout;
+	frame_title_layout_t title_layout;
 	int i;
 
 	if (!HAS_TITLE(fw))
@@ -322,7 +320,7 @@ static void frame_setup_titlebar(
 	}
 	frame_get_titlebar_dimensions(fw, frame_g, diff_g, &title_layout);
 	/* configure buttons */
-	for (i = 0; i < NUMBER_OF_BUTTONS; i++)
+	for (i = 0; i < NUMBER_OF_TITLE_BUTTONS; i++)
 	{
 		if (FW_W_BUTTON(fw, i) != None && (setup_parts & PART_BUTTONS))
 		{
@@ -396,6 +394,7 @@ static void __frame_setup_window(
 		mr_args = frame_create_move_resize_args(
 			fw, mode, NULL, &new_g, 0, DIR_NONE);
 		frame_move_resize(fw, mr_args);
+		((mr_args_internal *)mr_args)->flags.was_moved = 0;
 		frame_free_move_resize_args(fw, mr_args);
 		fw->frame_g = *frame_g;
 	}
@@ -519,20 +518,20 @@ static int frame_get_titlebar_compression(
  *   The mode for the resize operation
  */
 static void frame_get_resize_decor_gravities(
-	frame_decor_gravities_type *ret_grav, direction_type title_dir,
+	frame_decor_gravities_type *ret_grav, direction_t title_dir,
 	frame_move_resize_mode rmode, rectangle *delta_g)
 {
 	frame_decor_gravities_type grav_x;
 	frame_decor_gravities_type grav_y;
-	direction_type title_dir_x;
-	direction_type title_dir_y;
+	direction_t title_dir_x;
+	direction_t title_dir_y;
 
 	gravity_split_xy_dir(&title_dir_x, &title_dir_y, title_dir);
 	get_resize_decor_gravities_one_axis(
 		&grav_x, title_dir_x, rmode, DIR_W, DIR_E, (delta_g->x != 0));
 	get_resize_decor_gravities_one_axis(
 		&grav_y, title_dir_y, rmode, DIR_N, DIR_S, (delta_g->y != 0));
-	combine_decor_gravities(ret_grav, &grav_x, &grav_y);
+	combine_gravities(ret_grav, &grav_x, &grav_y);
 
 	return;
 }
@@ -592,7 +591,7 @@ static void frame_set_decor_gravities(
 	{
 		xcwa.win_gravity = grav->lbutton_grav;
 	}
-	for (i = 0; i < NUMBER_OF_BUTTONS; i += 2)
+	for (i = 0; i < NUMBER_OF_TITLE_BUTTONS; i += 2)
 	{
 		if (FW_W_BUTTON(fw, i))
 		{
@@ -608,7 +607,7 @@ static void frame_set_decor_gravities(
 	{
 		xcwa.win_gravity = grav->rbutton_grav;
 	}
-	for (i = 1; i < NUMBER_OF_BUTTONS; i += 2)
+	for (i = 1; i < NUMBER_OF_TITLE_BUTTONS; i += 2)
 	{
 		if (FW_W_BUTTON(fw, i))
 		{
@@ -631,7 +630,7 @@ static void frame_restore_client_gravities(FvwmWindow *fw)
 	{
 		fw->attr_backup.is_bit_gravity_stored = 0;
 		xcwa.bit_gravity = fw->attr_backup.bit_gravity;
-		valuemask = CWBitGravity;
+		valuemask |= CWBitGravity;
 	}
 	xcwa.win_gravity = fw->hints.win_gravity;
 	XChangeWindowAttributes(dpy, FW_W(fw), valuemask, &xcwa);
@@ -659,8 +658,8 @@ static rectangle *frame_get_hidden_pos(
 	rectangle *ret_hidden_g)
 {
 	rectangle *target_g;
-	direction_type dir_x;
-	direction_type dir_y;
+	direction_t dir_x;
+	direction_t dir_y;
 
 	if (do_unhide == False)
 	{
@@ -711,6 +710,7 @@ static void frame_update_hidden_window_pos(
 	XMoveResizeWindow(
 		dpy, FW_W(fw), hidden_g.x, hidden_g.y, mra->client_g.width,
 		mra->client_g.height);
+	mra->flags.was_moved = 1;
 
 	return;
 }
@@ -916,6 +916,26 @@ static void frame_mrs_hide_unhide_parent(
 	return;
 }
 
+static void frame_mrs_hide_unhide_parent2(
+	FvwmWindow *fw, mr_args_internal *mra)
+{
+	XSetWindowAttributes xswa;
+
+	/* finish hiding the parent */
+	if (mra->step_flags.do_hide_parent)
+	{
+		xswa.win_gravity = mra->grav.parent_grav;
+		XChangeWindowAttributes(
+			dpy, FW_W_PARENT(fw), CWWinGravity, &xswa);
+		/* update the hidden position of the client */
+		frame_update_hidden_window_pos(fw, mra, False);
+		XLowerWindow(dpy, FW_W_PARENT(fw));
+		XMapWindow(dpy, FW_W_PARENT(fw));
+	}
+
+	return;
+}
+
 static void frame_mrs_setup_draw_decorations(
 	FvwmWindow *fw, mr_args_internal *mra)
 {
@@ -984,11 +1004,14 @@ static void frame_mrs_resize_move_windows(
 			XMoveResizeWindow(
 				dpy, FW_W(fw), 0, 0, mra->parent_s.width,
 				mra->parent_s.height);
+			mra->client_g.width = mra->parent_s.width;
+			mra->client_g.height = mra->parent_s.height;
 		}
 		else
 		{
 			XMoveWindow(dpy, FW_W(fw), 0, 0);
 		}
+		mra->flags.was_moved = 1;
 #if 0
 		/* reduces flickering */
 		/* dv (11-Aug-2002): ... and slows down large scripts
@@ -1074,25 +1097,6 @@ static void frame_mrs_resize_move_windows(
 	return;
 }
 
-static void frame_mrs_hide_unhide_parent2(
-	FvwmWindow *fw, mr_args_internal *mra)
-{
-	XSetWindowAttributes xswa;
-
-	/* finish hiding the parent */
-	if (mra->step_flags.do_hide_parent)
-	{
-		xswa.win_gravity = mra->grav.parent_grav;
-		XChangeWindowAttributes(
-			dpy, FW_W_PARENT(fw), CWWinGravity, &xswa);
-		/* update the hidden position of the client */
-		frame_update_hidden_window_pos(fw, mra, False);
-		XLowerWindow(dpy, FW_W_PARENT(fw));
-		XMapWindow(dpy, FW_W_PARENT(fw));
-	}
-
-	return;
-}
 static void frame_move_resize_step(
 	FvwmWindow *fw, mr_args_internal *mra)
 {
@@ -1179,7 +1183,69 @@ static int frame_get_shading_laziness(
 	}
 }
 
-/* ---------------------------- interface functions ------------------------- */
+static void frame_reshape_border(FvwmWindow *fw)
+{
+	int dw;
+	int dh;
+	int grav;
+	int off_x = 0;
+	int off_y = 0;
+	rectangle naked_g;
+	rectangle *new_g;
+	size_borders b_old;
+	size_borders b_new;
+
+	get_window_borders(fw, &b_old);
+	dw = b_new.total_size.width - b_old.total_size.width;
+	dh = b_new.total_size.height - b_old.total_size.height;
+	/* calculate the new offsets */
+	if (!IS_MAXIMIZED(fw))
+	{
+		grav = fw->hints.win_gravity;
+		new_g = &fw->normal_g;
+	}
+	else
+	{
+		/* maximized windows are always considered to have
+		 * NorthWestGravity */
+		grav = NorthWestGravity;
+		new_g = &fw->max_g;
+		off_x = fw->normal_g.x - fw->max_g.x;
+		off_y = fw->normal_g.y - fw->max_g.y;
+	}
+	gravity_get_naked_geometry(grav, fw, &naked_g, new_g);
+	gravity_translate_to_northwest_geometry_no_bw(
+		grav, fw, &naked_g, &naked_g);
+	set_window_border_size(fw, fw->unshaped_boundary_width);
+	get_window_borders(fw, &b_new);
+	gravity_add_decoration(grav, fw, new_g, &naked_g);
+	if (IS_MAXIMIZED(fw))
+	{
+		/* prevent random paging when unmaximizing after the border
+		 * width has changed */
+		fw->max_offset.x += fw->normal_g.x - fw->max_g.x - off_x;
+		fw->max_offset.y += fw->normal_g.y - fw->max_g.y - off_y;
+	}
+	if (IS_SHADED(fw))
+	{
+		get_unshaded_geometry(fw, new_g);
+		if (USED_TITLE_DIR_FOR_SHADING(fw))
+		{
+			SET_SHADED_DIR(fw, GET_TITLE_DIR(fw));
+		}
+		get_shaded_geometry(fw, &fw->frame_g, new_g);
+	}
+	else
+	{
+		get_relative_geometry(&fw->frame_g, new_g);
+	}
+	frame_force_setup_window(
+		fw, new_g->x, new_g->y, new_g->width, new_g->height, True);
+
+	return;
+}
+
+/* ---------------------------- interface functions ------------------------ */
 
 /* Initialise structures local to frame.c */
 void frame_init(void)
@@ -1248,7 +1314,7 @@ void frame_destroyed_frame(
 
 void frame_get_titlebar_dimensions(
 	FvwmWindow *fw, rectangle *frame_g, rectangle *diff_g,
-	frame_title_layout_type *title_layout)
+	frame_title_layout_t *title_layout)
 {
 	size_borders b;
 	int i;
@@ -1373,10 +1439,10 @@ void frame_get_titlebar_dimensions(
 	}
 
 	/* configure left buttons */
-	for (i = 0; i < NUMBER_OF_BUTTONS; i++)
+	for (i = 0; i < NUMBER_OF_TITLE_BUTTONS; i++)
 	{
-		if ((revert_button && !(i & 1)) || (!revert_button && (i & 1)) ||
-		    FW_W_BUTTON(fw, i) == None)
+		if ((revert_button && !(i & 1)) ||
+		    (!revert_button && (i & 1)) || FW_W_BUTTON(fw, i) == None)
 		{
 			continue;
 		}
@@ -1420,10 +1486,10 @@ void frame_get_titlebar_dimensions(
 	*padd_coord += t_length;
 	/* configure right buttons */
 	*padd_coord -= br_sub;
-	for (i = NUMBER_OF_BUTTONS-1; i > -1; i--)
+	for (i = NUMBER_OF_TITLE_BUTTONS-1; i > -1; i--)
 	{
-		if ((revert_button && (i & 1)) || (!revert_button && !(i & 1)) ||
-		    FW_W_BUTTON(fw, i) == None)
+		if ((revert_button && (i & 1)) ||
+		    (!revert_button && !(i & 1)) || FW_W_BUTTON(fw, i) == None)
 		{
 			continue;
 		}
@@ -1595,7 +1661,7 @@ int frame_window_id_to_context(
 	{
 		int i;
 
-		for (i = 0; i < NUMBER_OF_BUTTONS; i++)
+		for (i = 0; i < NUMBER_OF_TITLE_BUTTONS; i++)
 		{
 			if (w == FW_W_BUTTON(fw, i) &&
 			    ((!(i & 1) && i / 2 < Scr.nr_left_buttons) ||
@@ -1649,6 +1715,7 @@ frame_move_resize_args frame_create_move_resize_args(
 {
 	mr_args_internal *mra;
 	Bool dummy;
+	Bool rc;
 	int whdiff;
 	int xydiff;
 	int diff;
@@ -1680,11 +1747,26 @@ frame_move_resize_args frame_create_move_resize_args(
 		mra->flags.do_not_configure_client = 1;
 	}
 	mra->mode = mr_mode;
-	mra->shade_dir = (direction_type)shade_dir;
+	mra->shade_dir = (direction_t)shade_dir;
 	mra->w_with_focus = (fw == get_focus_window()) ? FW_W(fw) : None;
 
 	/* calculate various geometries */
-	get_client_geometry(fw, &mra->client_g);
+	rc = XGetGeometry(
+		dpy, FW_W(fw), &JunkRoot, &mra->client_g.x, &mra->client_g.y,
+		&mra->client_g.width, &mra->client_g.height, &JunkBW,
+		&JunkDepth);
+	if (rc == True)
+	{
+		rc = XTranslateCoordinates(
+			dpy, FW_W_PARENT(fw), Scr.Root, mra->client_g.x,
+			mra->client_g.y, &mra->client_g.x, &mra->client_g.y,
+			&JunkChild);
+	}
+	if (rc == False)
+	{
+		/* Can only happen if the window died */
+		get_client_geometry(fw, &mra->client_g);
+	}
 	get_window_borders(fw, &mra->b_g);
 	get_window_borders_no_title(fw, &mra->b_no_title_g);
 	mra->start_g = (start_g != NULL) ? *start_g : fw->frame_g;
@@ -1733,6 +1815,7 @@ frame_move_resize_args frame_create_move_resize_args(
 	}
 
 	/* various flags */
+	mra->flags.was_moved = 0;
 	mra->step_flags.was_hidden =
 		(frame_is_parent_hidden(fw, &mra->start_g) == True);
 	mra->flags.is_setup =
@@ -1849,7 +1932,7 @@ void frame_free_move_resize_args(
 			dpy, FW_W_FRAME(fw), FShapeBounding, 0, 0, None,
 			FShapeSet);
 	}
-	frame_setup_shape(fw, mra->end_g.width, mra->end_g.height);
+	frame_setup_shape(fw, mra->end_g.width, mra->end_g.height, fw->wShaped);
 	if (mra->flags.do_restore_gravity)
 	{
 		mra->grav.client_grav = fw->hints.win_gravity;
@@ -1861,7 +1944,13 @@ void frame_free_move_resize_args(
 	{
 		frame_restore_client_gravities(fw);
 	}
-
+	if (!IS_SHADED(fw) && mra->flags.was_moved)
+	{
+		SendConfigureNotify(
+			fw, fw->frame_g.x, fw->frame_g.y, fw->frame_g.width,
+			fw->frame_g.height, 0, True);
+		mra->flags.was_moved = 0;
+	}
 	focus_grab_buttons_on_layer(fw->layer);
 	/* free the memory */
 	free(mr_args);
@@ -1930,8 +2019,7 @@ void frame_setup_window(
 	g.y = y;
 	g.width = w;
 	g.height = h;
-	__frame_setup_window(
-		fw, &g, do_send_configure_notify, False, False);
+	__frame_setup_window(fw, &g, do_send_configure_notify, False, False);
 
 	return;
 }
@@ -1946,8 +2034,7 @@ void frame_setup_window_app_request(
 	g.y = y;
 	g.width = w;
 	g.height = h;
-	__frame_setup_window(
-		fw, &g, do_send_configure_notify, False, True);
+	__frame_setup_window(fw, &g, do_send_configure_notify, False, True);
 
 	return;
 }
@@ -1962,8 +2049,7 @@ void frame_force_setup_window(
 	g.y = y;
 	g.width = w;
 	g.height = h;
-	__frame_setup_window(
-		fw, &g, do_send_configure_notify, True, False);
+	__frame_setup_window(fw, &g, do_send_configure_notify, True, False);
 
 	return;
 }
@@ -1973,7 +2059,7 @@ void frame_force_setup_window(
  * Sets up the shaped window borders
  *
  ****************************************************************************/
-void frame_setup_shape(FvwmWindow *fw, int w, int h)
+void frame_setup_shape(FvwmWindow *fw, int w, int h, int shape_mode)
 {
 	XRectangle rect;
 	rectangle r;
@@ -1983,35 +2069,42 @@ void frame_setup_shape(FvwmWindow *fw, int w, int h)
 	{
 		return;
 	}
-	if (!fw->wShaped)
+	if (fw->wShaped != shape_mode)
+	{
+		fw->wShaped = shape_mode;
+		frame_reshape_border(fw);
+	}
+	if (!shape_mode)
 	{
 		/* unset shape */
 		FShapeCombineMask(
 			dpy, FW_W_FRAME(fw), FShapeBounding, 0, 0, None,
 			FShapeSet);
-		return;
 	}
-	/* shape the window */
-	get_window_borders(fw, &b);
-	FShapeCombineShape(
-		dpy, FW_W_FRAME(fw), FShapeBounding, b.top_left.width,
-		b.top_left.height, FW_W(fw), FShapeBounding, FShapeSet);
-	if (FW_W_TITLE(fw))
+	else
 	{
-		/* windows w/ titles */
-		r.width = w;
-		r.height = h;
-		get_title_geometry(fw, &r);
-		rect.x = r.x;
-		rect.y = r.y;
-		rect.width = r.width;
-		rect.height = r.height;
-		FShapeCombineRectangles(
-			dpy, FW_W_FRAME(fw), FShapeBounding, 0, 0, &rect, 1,
-			FShapeUnion, Unsorted);
+		/* shape the window */
+		get_window_borders(fw, &b);
+		FShapeCombineShape(
+			dpy, FW_W_FRAME(fw), FShapeBounding, b.top_left.width,
+			b.top_left.height, FW_W(fw), FShapeBounding, FShapeSet);
+		if (FW_W_TITLE(fw))
+		{
+			/* windows w/ titles */
+			r.width = w;
+			r.height = h;
+			get_title_geometry(fw, &r);
+			rect.x = r.x;
+			rect.y = r.y;
+			rect.width = r.width;
+			rect.height = r.height;
+			FShapeCombineRectangles(
+				dpy, FW_W_FRAME(fw), FShapeBounding, 0, 0,
+				&rect, 1, FShapeUnion, Unsorted);
+		}
 	}
 
 	return;
 }
 
-/* ---------------------------- builtin commands ---------------------------- */
+/* ---------------------------- builtin commands --------------------------- */
