@@ -42,6 +42,7 @@
 #include "icons.h"
 #include "gnome.h"
 #include "ewmh.h"
+#include "frame.h"
 
 /* ---------------------------- local definitions -------------------------- */
 
@@ -57,8 +58,6 @@
 
 /* ---------------------------- imports ------------------------------------ */
 
-extern Bool is_frame_hide_window();
-
 /* ---------------------------- included code files ------------------------ */
 
 /* ---------------------------- local types -------------------------------- */
@@ -784,7 +783,7 @@ static void __raise_or_lower_window(
 		 * any style grabfocusoff override_redirect windows that may be
 		 * above it. This is used to cope with ill-behaved applications
 		 * that insist on using long-lived override_redirects. */
-		if (Scr.bo.RaiseOverUnmanaged)
+		if (Scr.bo.do_raise_over_unmanaged)
 		{
 			raise_over_unmanaged(t);
 		}
@@ -793,7 +792,7 @@ static void __raise_or_lower_window(
 		 * The following is a hack to raise X windows over native
 		 * windows which is needed for some (all ?) X servers running
 		 * under Windows or Windows NT. */
-		if (Scr.bo.RaiseHackNeeded)
+		if (Scr.bo.is_raise_hack_needed)
 		{
 			/* RBW - 09/20/1999. I find that trying to raise
 			 * unmanaged windows causes problems with some apps. If
@@ -1156,8 +1155,6 @@ static Bool is_above_unmanaged(FvwmWindow *fw, Window *umtop)
 	  Called from raise_over_unmanaged and is_on_top_of_layer.
 	*/
 	Bool ontop = True;
-
-
 	Window junk;
 	Window *tops;
 	int i;
@@ -1165,6 +1162,10 @@ static Bool is_above_unmanaged(FvwmWindow *fw, Window *umtop)
 	Window OR_Above = None;
 	XWindowAttributes wa;
 
+	if (fw->Desk != Scr.CurrentDesk)
+	{
+		return True;
+	}
 	if (!XQueryTree(dpy, Scr.Root, &junk, &junk, &tops, &num))
 	{
 		return ontop;
@@ -1193,8 +1194,7 @@ static Bool is_above_unmanaged(FvwmWindow *fw, Window *umtop)
 		if (wa.override_redirect == True
 		    && wa.class != InputOnly
 		    && tops[i] != Scr.NoFocusWin
-		    && (! is_frame_hide_window(tops[i]))
-			)
+		    && (!is_frame_hide_window(tops[i])))
 		{
 			OR_Above = tops[i];
 		}
@@ -1239,7 +1239,7 @@ static Bool is_on_top_of_layer_ignore_rom(FvwmWindow *fw)
 		 * on top by checking if the window overlaps another one.  If
 		 * it was below unmanaged windows, but on top of its layer, it
 		 * would be considered on top. */
-		if (Scr.bo.RaiseOverUnmanaged || overlap(fw, t))
+		if (Scr.bo.do_raise_over_unmanaged || overlap(fw, t))
 		{
 			if (!DO_RAISE_TRANSIENT(fw) ||
 			    (!IS_IN_TRANSIENT_SUBTREE(t) && t != fw))
@@ -1257,7 +1257,7 @@ static Bool __is_on_top_of_layer(FvwmWindow *fw, Bool client_entered)
 {
 	Window	junk;
 	Bool  ontop	= False;
-	if (Scr.bo.RaiseOverUnmanaged)
+	if (Scr.bo.do_raise_over_unmanaged)
 	{
 
 #define EXPERIMENTAL_ROU_HANDLING
