@@ -1,3 +1,4 @@
+/* -*-c-*- */
 /* This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
  * the Free Software Foundation; either version 2 of the License, or
@@ -13,18 +14,18 @@
  * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
  */
 
-/****************************************************************************
+/*
  * This module is all new
  * by Rob Nation
  * A little of it is borrowed from ctwm.
  * Copyright 1993 Robert Nation. No restrictions are placed on this code,
  * as long as the copyright notice is preserved
- ****************************************************************************/
-/***********************************************************************
+ */
+/*
  *
  * fvwm window-list popup code
  *
- ***********************************************************************/
+ */
 
 #include "config.h"
 
@@ -32,16 +33,13 @@
 #include <limits.h>
 
 #include "libs/fvwmlib.h"
-#include <libs/gravity.h>
-#include <libs/FScreen.h>
+#include "libs/FScreen.h"
+#include "libs/FGettext.h"
 #include "fvwm.h"
 #include "externs.h"
-#include "cursor.h"
 #include "functions.h"
-#include "bindings.h"
 #include "misc.h"
 #include "screen.h"
-#include "fvwmlib.h"
 #include "menus.h"
 #include "conditional.h"
 #include "stack.h"
@@ -49,29 +47,29 @@
 #include "virtual.h"
 #include "geometry.h"
 
-extern FvwmWindow *Fw;
-
 #define SHOW_GEOMETRY		(1<<0)
 #define SHOW_ALLDESKS		(1<<1)
 #define SHOW_NORMAL		(1<<2)
 #define SHOW_ICONIC		(1<<3)
-#define SHOW_STICKY		(1<<4)
+#define SHOW_STICKY_ACROSS_PAGES (1<<4)
+#define SHOW_STICKY_ACROSS_DESKS (1<<5)
 #define NO_DESK_SORT		(1<<6)
 #define SHOW_ICONNAME		(1<<7)
 #define SHOW_ALPHABETIC		(1<<8)
-#define SORT_CLASSNAME		(1<<9)
-#define SORT_REVERSE		(1<<10)
-#define SHOW_INFONOTGEO		(1<<11)
-#define NO_DESK_NUM		(1<<12)
-#define NO_CURRENT_DESK_TITLE	(1<<13)
-#define TITLE_FOR_ALL_DESKS	(1<<14)
-#define NO_NUM_IN_DESK_TITLE	(1<<15)
-#define SHOW_PAGE_X		(1<<16)
-#define SHOW_PAGE_Y		(1<<17)
-#define NO_LAYER		(1<<18)
-#define SHOW_SCREEN		(1<<19)
+#define SORT_BYCLASS		(1<<9)
+#define SORT_BYRESOURCE		(1<<10)
+#define SORT_REVERSE		(1<<11)
+#define SHOW_INFONOTGEO		(1<<12)
+#define NO_DESK_NUM		(1<<13)
+#define NO_CURRENT_DESK_TITLE	(1<<14)
+#define TITLE_FOR_ALL_DESKS	(1<<15)
+#define NO_NUM_IN_DESK_TITLE	(1<<16)
+#define SHOW_PAGE_X		(1<<17)
+#define SHOW_PAGE_Y		(1<<18)
+#define NO_LAYER		(1<<19)
+#define SHOW_SCREEN		(1<<20)
 #define SHOW_DEFAULT (SHOW_GEOMETRY | SHOW_ALLDESKS | SHOW_NORMAL | \
-	SHOW_ICONIC | SHOW_STICKY)
+	SHOW_ICONIC | SHOW_STICKY_ACROSS_PAGES | SHOW_STICKY_ACROSS_DESKS)
 
 static char *get_desk_title(int desk, unsigned long flags, Bool is_top_title)
 {
@@ -94,20 +92,20 @@ static char *get_desk_title(int desk, unsigned long flags, Bool is_top_title)
 		{
 			sprintf(tlabel, "%s%s", desk_name,
 				(is_top_title && (flags & SHOW_GEOMETRY)) ?
-				"\tGeometry" : "");
+				_("\tGeometry") : "");
 		}
 		else
 		{
 			sprintf(tlabel,"%d: %s%s", desk, desk_name,
 				(is_top_title && (flags & SHOW_GEOMETRY)) ?
-				"\tGeometry" : "");
+				_("\tGeometry") : "");
 		}
 	}
 	else
 	{
-		sprintf(tlabel,"Desk: %d%s",desk,
+		sprintf(tlabel,_("Desk: %d%s"),desk,
 			(is_top_title && (flags & SHOW_GEOMETRY)) ?
-			"\tGeometry" : "");
+			_("\tGeometry") : "");
 	}
 
 	return tlabel;
@@ -130,7 +128,23 @@ static int (*compare)(const FvwmWindow **a, const FvwmWindow **b);
 static int classCompare(const FvwmWindow **a, const FvwmWindow **b)
 {
 	int result = strcasecmp((*a)->class.res_class, (*b)->class.res_class);
+	if (result)
+	{
+		return result;
+	}
 
+	return strcasecmp((*a)->visible_name, (*b)->visible_name);
+}
+
+static int resourceCompare(const FvwmWindow **a, const FvwmWindow **b)
+{
+	int result = strcasecmp((*a)->class.res_class, (*b)->class.res_class);
+	if (result)
+	{
+		return result;
+	}
+
+	result = strcasecmp((*a)->class.res_name, (*b)->class.res_name);
 	if (result)
 	{
 		return result;
@@ -141,8 +155,24 @@ static int classCompare(const FvwmWindow **a, const FvwmWindow **b)
 
 static int classIconCompare(const FvwmWindow **a, const FvwmWindow **b)
 {
-	int result = strcasecmp((*a)->class.res_class,
-				(*b)->class.res_class);
+	int result = strcasecmp((*a)->class.res_class, (*b)->class.res_class);
+	if (result)
+	{
+		return result;
+	}
+
+	return strcasecmp((*a)->visible_icon_name, (*b)->visible_icon_name);
+}
+
+static int resourceIconCompare(const FvwmWindow **a, const FvwmWindow **b)
+{
+	int result = strcasecmp((*a)->class.res_class, (*b)->class.res_class);
+	if (result)
+	{
+		return result;
+	}
+
+	result = strcasecmp((*a)->class.res_name, (*b)->class.res_name);
 	if (result)
 	{
 		return result;
@@ -174,6 +204,7 @@ void CMD_WindowList(F_CMD_ARGS)
 	char tname[128];
 	char loc[64];
 	char *name=NULL;
+	Bool free_name = False;
 	int dwidth;
 	int dheight;
 	char *tlabel;
@@ -187,15 +218,14 @@ void CMD_WindowList(F_CMD_ARGS)
 	int desk = Scr.CurrentDesk;
 	unsigned long flags = SHOW_DEFAULT;
 	char *func = NULL;
+	char *ffunc = NULL;
 	char *tfunc = NULL;
 	char *default_action = NULL;
 	MenuReturn mret;
-	XEvent *teventp;
 	MenuOptions mops;
 	int low_layer = 0;  /* show all layers by default */
 	int high_layer = INT_MAX;
 	int max_label_width = 0;
-	int tc;
 	int show_listskip = 0; /* do not show listskip by default */
 	Bool use_hotkey = True;
 	KeyCode old_sor_keycode;
@@ -212,6 +242,9 @@ void CMD_WindowList(F_CMD_ARGS)
 	Bool first_desk = True;
 	Bool empty_menu = True;
 	Bool was_get_menu_opts_called = False;
+	FvwmWindow * const fw = exc->w.fw;
+	const Window w = exc->w.w;
+	const exec_context_t *exc2;
 
 	memset(&mops, 0, sizeof(mops));
 	memset(&mret, 0, sizeof(MenuReturn));
@@ -236,7 +269,7 @@ void CMD_WindowList(F_CMD_ARGS)
 			free(cond_flags);
 		}
 		opts = get_menu_options(
-			action, w, fw, eventp, NULL, NULL, &mops);
+			action, w, fw, NULL, NULL, NULL, &mops);
 		was_get_menu_opts_called = True;
 
 		/* parse options */
@@ -279,9 +312,13 @@ void CMD_WindowList(F_CMD_ARGS)
 			{
 				flags |= SHOW_ALPHABETIC;
 			}
-			else if (StrEquals(tok,"SortClassName"))
+			else if (StrEquals(tok,"SortByClass"))
 			{
-				flags |= SORT_CLASSNAME;
+				flags |= SORT_BYCLASS;
+			}
+			else if (StrEquals(tok,"SortByResource"))
+			{
+				flags |= SORT_BYRESOURCE;
 			}
 			else if (StrEquals(tok,"ReverseOrder"))
 			{
@@ -360,15 +397,42 @@ void CMD_WindowList(F_CMD_ARGS)
 			}
 			else if (StrEquals(tok,"NoSticky"))
 			{
-				flags &= ~SHOW_STICKY;
+				flags &= ~(SHOW_STICKY_ACROSS_PAGES);
+				flags &= ~(SHOW_STICKY_ACROSS_DESKS);
+			}
+			else if (StrEquals(tok,"NoStickyPage"))
+			{
+				flags &= ~(SHOW_STICKY_ACROSS_PAGES);
+			}
+			else if (StrEquals(tok,"NoStickyDesk"))
+			{
+				flags &= ~(SHOW_STICKY_ACROSS_DESKS);
 			}
 			else if (StrEquals(tok,"Sticky"))
 			{
-				flags |= SHOW_STICKY;
+				flags |= SHOW_STICKY_ACROSS_PAGES;
+				flags |= SHOW_STICKY_ACROSS_DESKS;
+			}
+			else if (StrEquals(tok,"StickyPage"))
+			{
+				flags |= SHOW_STICKY_ACROSS_PAGES;
+			}
+			else if (StrEquals(tok,"StickyDesk"))
+			{
+				flags |= SHOW_STICKY_ACROSS_DESKS;
 			}
 			else if (StrEquals(tok,"OnlySticky"))
 			{
-				flags = SHOW_STICKY;
+				flags = SHOW_STICKY_ACROSS_PAGES;
+				flags = SHOW_STICKY_ACROSS_DESKS;
+			}
+			else if (StrEquals(tok,"OnlyStickyPage"))
+			{
+				flags = SHOW_STICKY_ACROSS_PAGES;
+			}
+			else if (StrEquals(tok,"OnlyStickyDesk"))
+			{
+				flags = SHOW_STICKY_ACROSS_DESKS;
 			}
 			else if (StrEquals(tok,"UseListSkip"))
 			{
@@ -494,7 +558,7 @@ void CMD_WindowList(F_CMD_ARGS)
 	if (was_get_menu_opts_called == False)
 	{
 		opts = get_menu_options(
-			action, w, fw, eventp, NULL, NULL, &mops);
+			action, w, fw, NULL, NULL, NULL, &mops);
 	}
 
 	tlabel = get_desk_title(desk, flags, True);
@@ -577,33 +641,46 @@ void CMD_WindowList(F_CMD_ARGS)
 	}
 
 	/* Do alphabetic sort */
-	if (flags & (SHOW_ALPHABETIC | SORT_CLASSNAME))
+	if (flags & (SHOW_ALPHABETIC | SORT_BYCLASS | SORT_BYRESOURCE))
 	{
 		/* This will be compare or compareReverse if a reverse order
 		 * is selected. */
 		int (*sort)(const FvwmWindow **a, const FvwmWindow **b);
 
-		switch(flags & (SHOW_ALPHABETIC|SHOW_ICONNAME|SORT_CLASSNAME))
+		switch (flags & (SHOW_ALPHABETIC | SHOW_ICONNAME | \
+			SORT_BYCLASS | SORT_BYRESOURCE))
 		{
 		case SHOW_ALPHABETIC:
 			compare = visibleCompare;
 			break;
-		case SHOW_ALPHABETIC|SHOW_ICONNAME:
+		case SHOW_ALPHABETIC | SHOW_ICONNAME:
 			compare = iconCompare;
 			break;
-			/* Sorting based on class name produces an alphabetic
-			 * order so the keyword alphabetic is redundant. */
-		case SORT_CLASSNAME:
-		case SORT_CLASSNAME|SHOW_ALPHABETIC:
+		/* Sorting based on class name produces an alphabetic
+		 * order so the keyword alphabetic is redundant. */
+		case SORT_BYCLASS:
+		case SORT_BYCLASS | SHOW_ALPHABETIC:
 			compare = classCompare;
 			break;
-		case SORT_CLASSNAME|SHOW_ICONNAME:
-		case SORT_CLASSNAME|SHOW_ALPHABETIC|SHOW_ICONNAME:
+		case SORT_BYCLASS | SHOW_ICONNAME:
+		case SORT_BYCLASS | SHOW_ICONNAME | SHOW_ALPHABETIC:
 			compare = classIconCompare;
 			break;
-			/* All current cases are covered, but if something
-			 * changes in the future we leave compare valid even if
-			 * it isn't what is expected. */
+		case SORT_BYRESOURCE:
+		case SORT_BYRESOURCE | SORT_BYCLASS:
+		case SORT_BYRESOURCE | SORT_BYCLASS | SHOW_ALPHABETIC:
+			compare = resourceCompare;
+			break;
+		case SORT_BYRESOURCE | SHOW_ICONNAME:
+		case SORT_BYRESOURCE | SHOW_ICONNAME | SORT_BYCLASS:
+		case SORT_BYRESOURCE | SHOW_ICONNAME | SORT_BYCLASS | \
+		SHOW_ALPHABETIC:
+			compare = resourceIconCompare;
+			break;
+
+		/* All current cases are covered, but if something
+		 * changes in the future we leave compare valid even if
+		 * it isn't what is expected. */
 		default:
 			compare = visibleCompare;
 			break;
@@ -654,292 +731,295 @@ void CMD_WindowList(F_CMD_ARGS)
 		for (ii = 0; ii < numWindows; ii++)
 		{
 			t = windowList[ii];
-			if (((t->Desk == next_desk) || (flags & NO_DESK_SORT)))
+			if (t->Desk != next_desk && !(flags & NO_DESK_SORT))
 			{
-				if (!show_listskip && DO_SKIP_WINDOW_LIST(t))
-				{
-					/* don't want listskip windows - skip */
-					continue;
-				}
-				if (show_listskip == 2 &&
-				    !DO_SKIP_WINDOW_LIST(t))
-				{
-					/* don't want no listskip one - skip */
-					continue;
-				}
-				if (use_condition &&
-				    !MatchesConditionMask(t, &mask))
-				{
-					/* doesn't match specified condition */
-					continue;
-				}
-				if (!(flags & SHOW_ICONIC) && (IS_ICONIFIED(t)))
-				{
-					/* don't want icons - skip */
-					continue;
-				}
-				if (!(flags & SHOW_STICKY) && (IS_STICKY(t)))
-				{
-					/* don't want sticky ones - skip */
-					continue;
-				}
-				if (!(flags & SHOW_NORMAL) &&
-				    !((IS_ICONIFIED(t)) || (IS_STICKY(t))))
-				{
-					/* don't want "normal" ones - skip */
-					continue;
-				}
-				if (get_layer(t) < low_layer ||
-				    get_layer(t) > high_layer)
-				{
-					/* don't want this layer */
-					continue;
-				}
+				continue;
+			}
+			if (!show_listskip && DO_SKIP_WINDOW_LIST(t))
+			{
+				/* don't want listskip windows - skip */
+				continue;
+			}
+			if (show_listskip == 2 &&
+			    !DO_SKIP_WINDOW_LIST(t))
+			{
+				/* don't want no listskip one - skip */
+				continue;
+			}
+			if (use_condition && !MatchesConditionMask(t, &mask))
+			{
+				/* doesn't match specified condition */
+				continue;
+			}
+			if (!(flags & SHOW_ICONIC) && (IS_ICONIFIED(t)))
+			{
+				/* don't want icons - skip */
+				continue;
+			}
+			if (!(flags & SHOW_STICKY_ACROSS_PAGES) &&
+			    (IS_STICKY_ACROSS_PAGES(t)))
+			{
+				/* don't want sticky ones - skip */
+				continue;
+			}
+			if (!(flags & SHOW_STICKY_ACROSS_DESKS) &&
+			    (IS_STICKY_ACROSS_DESKS(t)))
+			{
+				/* don't want sticky ones - skip */
+				continue;
+			}
+			if (!(flags & SHOW_NORMAL) &&
+			    !(IS_ICONIFIED(t) ||
+			      IS_STICKY_ACROSS_PAGES(t) ||
+			      IS_STICKY_ACROSS_DESKS(t)))
+			{
+				/* don't want "normal" ones - skip */
+				continue;
+			}
+			if (get_layer(t) < low_layer ||
+			    get_layer(t) > high_layer)
+			{
+				/* don't want this layer */
+				continue;
+			}
 
-				empty_menu = False;
-				/* add separator between desks when geometry
-				 * shown but not at the top*/
-				if (t->Desk != last_desk_displayed)
+			empty_menu = False;
+			/* add separator between desks when geometry
+			 * shown but not at the top*/
+			if (t->Desk != last_desk_displayed)
+			{
+				if (last_desk_displayed != INT_MIN)
 				{
-					if (last_desk_displayed != INT_MIN)
+					if (((flags & SHOW_GEOMETRY) ||
+					     (flags & SHOW_INFONOTGEO)) &&
+					    !(flags & TITLE_FOR_ALL_DESKS))
 					{
-						if (((flags & SHOW_GEOMETRY) ||
-						     (flags &
-						      SHOW_INFONOTGEO)) &&
-						    !(flags &
-						      TITLE_FOR_ALL_DESKS))
-						{
-							AddToMenu(
-								mr, NULL, NULL,
-								False, False,
-								False);
-						}
-						if (flags & TITLE_FOR_ALL_DESKS)
-						{
-							tlabel = get_desk_title(
-								t->Desk, flags,
-								False);
-							AddToMenu(
-								mr, tlabel,
-								"TITLE", False,
-								False, False);
-							free(tlabel);
-						}
+						AddToMenu(
+							mr, NULL, NULL, False,
+							False, False);
 					}
-					last_desk_displayed = t->Desk;
-				}
-				if (first_desk && flags & TITLE_FOR_ALL_DESKS)
-				{
-					tlabel = get_desk_title(
-						t->Desk, flags, False);
-					AddToMenu(
-						mr, tlabel, "TITLE", False,
-						False, False);
-					free(tlabel);
-				}
-				first_desk = False;
-
-				if (flags & SHOW_ICONNAME)
-				{
-					name = t->visible_icon_name;
-				}
-				else
-				{
-					name = t->visible_name;
-				}
-
-				if (!name)
-				{
-					name = "NULL_NAME";
-				}
-				else if (max_label_width > 0 &&
-					 strlen(name) > max_label_width)
-				{
-					name = strdup(name);
-					name[max_label_width] = '\0';
-				}
-
-				t_hot = safemalloc(strlen(name) + 80);
-				if (use_hotkey)
-				{
-					/* Generate label */
-					sprintf(t_hot, "&%c. ", scut);
-				}
-				else
-				{
-					*t_hot = 0;
-				}
-				if (!(flags & SHOW_INFONOTGEO))
-				{
-					strcat(t_hot, name);
-				}
-				if (*t_hot == 0)
-				{
-					strcpy(t_hot, " ");
-				}
-
-				/* Next shortcut key */
-				if (scut == '9')
-				{
-					scut = 'A';
-				}
-				else if (scut == 'Z')
-				{
-					scut = '0';
-				}
-				else
-				{
-					scut++;
-				}
-
-				if (flags & SHOW_INFONOTGEO)
-				{
-					tname[0]=0;
-					if (!IS_ICONIFIED(t) &&
-					    !(flags & NO_DESK_NUM))
+					if (flags & TITLE_FOR_ALL_DESKS)
 					{
-						sprintf(loc,"%d:", t->Desk);
-						strcat(tname,loc);
-					}
-					if (IS_ICONIFIED(t))
-					{
-						strcat(tname, "(");
-					}
-					strcat(t_hot,"\t");
-					strcat(t_hot,tname);
-					strcat(t_hot, name);
-					if (IS_ICONIFIED(t))
-					{
-						strcat(t_hot, ")");
+						tlabel = get_desk_title(
+							t->Desk, flags, False);
+						AddToMenu(
+							mr, tlabel, "TITLE",
+							False, False, False);
+						free(tlabel);
 					}
 				}
-				else if (flags & SHOW_GEOMETRY)
-				{
-					size_borders b;
-
-					tname[0]=0;
-					if (IS_ICONIFIED(t))
-					{
-						strcpy(tname, "(");
-						strcat(tname, loc);
-					}
-					if (!(flags & NO_DESK_NUM))
-					{
-						sprintf(loc, "%d", t->Desk);
-						strcat(tname, loc);
-					}
-					if (flags & SHOW_SCREEN)
-					{
-						fscreen_scr_arg fscr;
-						int scr;
-
-						fscr.xypos.x =
-							Scr.Vx + t->frame_g.x +
-							t->frame_g.width / 2;
-						fscr.xypos.y =
-							Scr.Vy + t->frame_g.y +
-							t->frame_g.height / 2;
-						scr = FScreenGetScrId(
-							&fscr, FSCREEN_XYPOS);
-						sprintf(loc, "@%d", scr);
-						strcat(tname, loc);
-					}
-					if (flags & SHOW_PAGE_X)
-					{
-						sprintf(loc, "+%d",
-							(Scr.Vx + t->frame_g.x +
-							 t->frame_g.width / 2) /
-							Scr.MyDisplayWidth);
-						strcat(tname, loc);
-					}
-					if (flags & SHOW_PAGE_Y)
-					{
-						sprintf(loc, "+%d",
-							(Scr.Vy + t->frame_g.y +
-							 t->frame_g.height/2) /
-							Scr.MyDisplayHeight);
-						strcat(tname, loc);
-					}
-					if (!(flags & NO_LAYER))
-					{
-						sprintf(loc, "(%d)",
-							(get_layer(t)));
-						strcat(tname, loc);
-					}
-					strcat(tname, ":");
-					get_window_borders(t, &b);
-					dheight = t->frame_g.height -
-						b.total_size.height;
-					dwidth = t->frame_g.width -
-						b.total_size.width;
-					dwidth -= t->hints.base_width;
-					dheight -= t->hints.base_height;
-					dwidth /= t->hints.width_inc;
-					dheight /= t->hints.height_inc;
-
-					sprintf(loc,"%d",dwidth);
-					strcat(tname, loc);
-					sprintf(loc,"x%d",dheight);
-					strcat(tname, loc);
-					if (t->frame_g.x >=0)
-					{
-						sprintf(loc,"+%d",t->frame_g.x);
-					}
-					else
-					{
-						sprintf(loc,"%d",t->frame_g.x);
-					}
-					strcat(tname, loc);
-					if (t->frame_g.y >=0)
-					{
-						sprintf(loc,"+%d",t->frame_g.y);
-					}
-					else
-					{
-						sprintf(loc,"%d",t->frame_g.y);
-					}
-					strcat(tname, loc);
-
-					if (IS_STICKY(t))
-					{
-						strcat(tname, " S");
-					}
-					if (IS_ICONIFIED(t))
-					{
-						strcat(tname, ")");
-					}
-					strcat(t_hot,"\t");
-					strcat(t_hot,tname);
-				}
-				if (!func)
-				{
-					tfunc = safemalloc(40);
-					sprintf(tfunc, "WindowListFunc %lu",
-						FW_W(t));
-				}
-				else
-				{
-					tfunc = safemalloc(strlen(func) + 32);
-					sprintf(tfunc,"%s %lu", func, FW_W(t));
-				}
+				last_desk_displayed = t->Desk;
+			}
+			if (first_desk && flags & TITLE_FOR_ALL_DESKS)
+			{
+				tlabel = get_desk_title(t->Desk, flags, False);
 				AddToMenu(
-					mr, t_hot, tfunc, False, False, False);
-				free(tfunc);
-				/* Add the title pixmap */
-				if (FMiniIconsSupported && t->mini_icon)
+					mr, tlabel, "TITLE", False, False,
+					False);
+				free(tlabel);
+			}
+			first_desk = False;
+
+			if (flags & SHOW_ICONNAME)
+			{
+				name = t->visible_icon_name;
+			}
+			else
+			{
+				name = t->visible_name;
+			}
+
+			free_name = False;
+			if (!name)
+			{
+				name = "NULL_NAME";
+			}
+			else if (max_label_width > 0 &&
+				 strlen(name) > max_label_width)
+			{
+				name = strdup(name);
+				name[max_label_width] = '\0';
+				free_name = True;
+			}
+
+			t_hot = safemalloc(strlen(name) + 80);
+			if (use_hotkey)
+			{
+				/* Generate label */
+				sprintf(t_hot, "&%c. ", scut);
+			}
+			else
+			{
+				*t_hot = 0;
+			}
+			if (!(flags & SHOW_INFONOTGEO))
+			{
+				strcat(t_hot, name);
+			}
+			if (*t_hot == 0)
+			{
+				strcpy(t_hot, " ");
+			}
+
+			/* Next shortcut key */
+			if (scut == '9')
+			{
+				scut = 'A';
+			}
+			else if (scut == 'Z')
+			{
+				scut = '0';
+			}
+			else
+			{
+				scut++;
+			}
+
+			if (flags & SHOW_INFONOTGEO)
+			{
+				tname[0]=0;
+				if (!IS_ICONIFIED(t) &&
+				    !(flags & NO_DESK_NUM))
 				{
-					MI_MINI_ICON(MR_LAST_ITEM(mr))[0] =
-						t->mini_icon;
-					/* increase the cache count. Otherwise
-					 * the pixmap will be eventually
-					 * removed from the cache by
-					 * DestroyMenu */
-					t->mini_icon->count++;
+					sprintf(loc,"%d:", t->Desk);
+					strcat(tname,loc);
 				}
-				if (t_hot)
+				if (IS_ICONIFIED(t))
 				{
-					free(t_hot);
+					strcat(tname, "(");
 				}
+				strcat(t_hot,"\t");
+				strcat(t_hot,tname);
+				strcat(t_hot, name);
+				if (IS_ICONIFIED(t))
+				{
+					strcat(t_hot, ")");
+				}
+			}
+			else if (flags & SHOW_GEOMETRY)
+			{
+				size_borders b;
+
+				tname[0]=0;
+				if (IS_ICONIFIED(t))
+				{
+					strcpy(tname, "(");
+				}
+				if (!(flags & NO_DESK_NUM))
+				{
+					sprintf(loc, "%d", t->Desk);
+					strcat(tname, loc);
+				}
+				if (flags & SHOW_SCREEN)
+				{
+					fscreen_scr_arg fscr;
+					int scr;
+
+					fscr.xypos.x =
+						Scr.Vx + t->frame_g.x +
+						t->frame_g.width / 2;
+					fscr.xypos.y =
+						Scr.Vy + t->frame_g.y +
+						t->frame_g.height / 2;
+					scr = FScreenGetScrId(
+						&fscr, FSCREEN_XYPOS);
+					sprintf(loc, "@%d", scr);
+					strcat(tname, loc);
+				}
+				if (flags & SHOW_PAGE_X)
+				{
+					sprintf(loc, "+%d",
+						(Scr.Vx + t->frame_g.x +
+						 t->frame_g.width / 2) /
+						Scr.MyDisplayWidth);
+					strcat(tname, loc);
+				}
+				if (flags & SHOW_PAGE_Y)
+				{
+					sprintf(loc, "+%d",
+						(Scr.Vy + t->frame_g.y +
+						 t->frame_g.height/2) /
+						Scr.MyDisplayHeight);
+					strcat(tname, loc);
+				}
+				if (!(flags & NO_LAYER))
+				{
+					sprintf(loc, "(%d)",
+						(get_layer(t)));
+					strcat(tname, loc);
+				}
+				strcat(tname, ":");
+				get_window_borders(t, &b);
+				dheight = t->frame_g.height -
+					b.total_size.height;
+				dwidth = t->frame_g.width -
+					b.total_size.width;
+				dwidth -= t->hints.base_width;
+				dheight -= t->hints.base_height;
+				dwidth /= t->hints.width_inc;
+				dheight /= t->hints.height_inc;
+
+				sprintf(loc,"%d",dwidth);
+				strcat(tname, loc);
+				sprintf(loc,"x%d",dheight);
+				strcat(tname, loc);
+				if (t->frame_g.x >=0)
+				{
+					sprintf(loc,"+%d",t->frame_g.x);
+				}
+				else
+				{
+					sprintf(loc,"%d",t->frame_g.x);
+				}
+				strcat(tname, loc);
+				if (t->frame_g.y >=0)
+				{
+					sprintf(loc,"+%d",t->frame_g.y);
+				}
+				else
+				{
+					sprintf(loc,"%d",t->frame_g.y);
+				}
+				strcat(tname, loc);
+
+				if (IS_STICKY_ACROSS_PAGES(t) ||
+				    IS_STICKY_ACROSS_DESKS(t))
+				{
+					strcat(tname, " S");
+				}
+				if (IS_ICONIFIED(t))
+				{
+					strcat(tname, ")");
+				}
+				strcat(t_hot,"\t");
+				strcat(t_hot,tname);
+			}
+			ffunc = func ? func : "WindowListFunc";
+			tfunc = safemalloc(strlen(ffunc) + 36);
+			/* support two ways for now: window context
+			 * (new) and window id param (old) */
+			sprintf(tfunc, "WindowId %lu %s %lu",
+				FW_W(t), ffunc, FW_W(t));
+			AddToMenu(
+				mr, t_hot, tfunc, False, False, False);
+			free(tfunc);
+			/* Add the title pixmap */
+			if (FMiniIconsSupported && t->mini_icon)
+			{
+				MI_MINI_ICON(MR_LAST_ITEM(mr))[0] =
+					t->mini_icon;
+				/* increase the cache count. Otherwise the
+				 * pixmap will be eventually removed from the
+				 * cache by DestroyMenu */
+				t->mini_icon->count++;
+			}
+			if (t_hot)
+			{
+				free(t_hot);
+			}
+			if (free_name)
+			{
+				free(name);
 			}
 		}
 	}
@@ -961,16 +1041,7 @@ void CMD_WindowList(F_CMD_ARGS)
 	{
 		free(iconifiedList);
 	}
-	if (!default_action && eventp && eventp->type == KeyPress)
-	{
-		teventp = (XEvent *)1;
-	}
-	else
-	{
-		teventp = eventp;
-	}
-
-	/* Use the WindowList menu style is there is one */
+	/* Use the WindowList menu style if there is one */
 	change_mr_menu_style(mr, "WindowList");
 
 	/* Activate select_on_release style */
@@ -979,13 +1050,6 @@ void CMD_WindowList(F_CMD_ARGS)
 	    (!MST_SELECT_ON_RELEASE_KEY(mr) ||
 	     sor_keyname != sor_default_keyname))
 	{
-		KeyCode keycode = 0;
-
-		if (sor_keyname)
-		{
-			keycode = XKeysymToKeycode(
-				dpy, FvwmStringToKeysym(dpy, sor_keyname));
-		}
 		MST_SELECT_ON_RELEASE_KEY(mr) =
 			XKeysymToKeycode(
 				dpy, FvwmStringToKeysym(dpy, sor_keyname));
@@ -993,18 +1057,16 @@ void CMD_WindowList(F_CMD_ARGS)
 
 	memset(&mp, 0, sizeof(mp));
 	mp.menu = mr;
-	t = Fw;
-	mp.pfw = &t;
-	tc = context;
-	mp.pcontext = &tc;
+	exc2 = exc_clone_context(exc, NULL, 0);
+	mp.pexc = &exc2;
 	mp.flags.has_default_action = (default_action && *default_action != 0);
-	mp.flags.is_sticky = True;
-	mp.flags.is_submenu = False;
-	mp.flags.is_already_mapped = False;
-	mp.eventp = teventp;
+	mp.flags.is_sticky = 1;
+	mp.flags.is_submenu = 0;
+	mp.flags.is_already_mapped = 0;
+	mp.flags.is_triggered_by_keypress =
+		(!default_action && exc->x.etrigger->type == KeyPress);
 	mp.pops = &mops;
 	mp.ret_paction = &ret_action;
-
 	do_menu(&mp, &mret);
 	/* Restore old menu style */
 	MST_SELECT_ON_RELEASE_KEY(mr) = old_sor_keycode;
@@ -1013,11 +1075,10 @@ void CMD_WindowList(F_CMD_ARGS)
 		free(ret_action);
 	}
 	DestroyMenu(mr, False, False);
-	if (mret.rc == MENU_DOUBLE_CLICKED && default_action && *default_action)
+	if (mret.rc == MENU_DOUBLE_CLICKED &&
+	    default_action && *default_action)
 	{
-		old_execute_function(
-			cond_rc, default_action, fw, eventp, context, *Module,
-			0, NULL);
+		execute_function(cond_rc, exc2, default_action, 0);
 	}
 	if (default_action != NULL)
 	{
@@ -1027,6 +1088,11 @@ void CMD_WindowList(F_CMD_ARGS)
 	{
 		FreeConditionMask(&mask);
 	}
+	if (sor_keyname && sor_keyname != sor_default_keyname)
+	{
+		free(sor_keyname);
+	}
+	exc_destroy_context(exc2);
 
 	return;
 }
