@@ -1,3 +1,4 @@
+/* -*-c-*- */
 /*
  * FvwmButtons, copyright 1996, Jarl Totland
  *
@@ -12,21 +13,6 @@
  *
  */
 
-/* This program is free software; you can redistribute it and/or modify
- * it under the terms of the GNU General Public License as published by
- * the Free Software Foundation; either version 2 of the License, or
- * (at your option) any later version.
- *
- * This program is distributed in the hope that it will be useful,
- * but WITHOUT ANY WARRANTY; without even the implied warranty of
- * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
- * GNU General Public License for more details.
- *
- * You should have received a copy of the GNU General Public License
- * along with this program; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
- */
-
 /* -------------------------------- DEBUG ---------------------------------- */
 
 /* Uncomment some of these defines to get (lots of) debug output. If you can't
@@ -87,6 +73,16 @@
 				 * button with a swallowed app. */
 #define b_ActionOnPress \
 		     0x02000000 /* By default this only done on Popup */
+#define b_Id             0x04000000 /* Has a user defined id for referencing */
+#define b_ActiveIcon     0x08000000 /* Use alternate Icon on hover */
+#define b_ActiveColorset 0x10000000 /* Use alternate colorset on hover */
+#define b_ActiveTitle    0x20000000 /* Use alternate Title text on hover */
+#define b_PressIcon      0x40000000 /* Use alternate Icon on press */
+#define b_PressColorset  0x80000000 /* Use alternate Colorset on press */
+/* FIXME: We're out of bits!
+   Nasty hack: b_PressColorset is used by UberButton & it would never use
+   b_PressTitle (& vice-versa) so they have the same bit-value. */
+#define b_PressTitle    0x80000000 /* Use alternate Title text on press */
 
 /* Flags for b->swallow */
 #define b_Count       0x0003 /* Init counter for swallowing */
@@ -133,6 +129,8 @@ struct container_info_struct
   char *back_file;         /* b_Back && b_IconBack */
   char *fore;              /* b_Fore */
   int colorset;            /* b_Colorset */
+  int activeColorset;      /* b_ActiveColorset */
+  int pressColorset;       /* b_PressColorset */
   Pixel fc;                /* b_Fore */
   Pixel bc,hc,sc;          /* b_Back && !b_IconBack */
   FvwmPicture *backicon;   /* b_Back && b_IconBack */
@@ -155,12 +153,13 @@ struct button_info_struct
 {
   /* required fields */
   unsigned long flags;
-  int  BPosX,BPosY;        /* position in button units from top left */
-  byte BWidth,BHeight;     /* width and height in button units  */
+  int  BPosX,BPosY;               /* position in button units from top left */
+  unsigned int BWidth,BHeight;     /* width and height in button units  */
   button_info *parent;
   int n;                   /* number in parent */
 
   /* conditional fields */ /* applicable if these flags are set */
+  char *id;                /* b_Id */
   FlocaleFont *Ffont;      /* b_Font */
   char *font_string;       /* b_Font */
   char *back;              /* b_Back */
@@ -172,15 +171,21 @@ struct button_info_struct
   byte justify_mask;       /* b_Justify */
   container_info *c;       /* b_Container */
   char *title;             /* b_Title */
+  char *activeTitle;       /* b_ActiveTitle */
+  char *pressTitle;        /* b_PressTitle */
   char **action;           /* b_Action */
   char *icon_file;         /* b_Icon */
+  char *active_icon_file;  /* b_ActiveIcon */
+  char *press_icon_file;   /* b_PressIcon */
   char *hangon;            /* b_Hangon || b_Swallow */
   Pixel fc;                /* b_Fore */
   Pixel bc,hc,sc;          /* b_Back && !b_IconBack */
   ushort minx,miny;        /* b_Size */
   FvwmPicture *icon;       /* b_Icon */
   FvwmPicture *backicon;   /* b_Back && b_IconBack */
-  Window IconWin;          /* b_Icon || b_Swallow */
+  FvwmPicture *activeicon; /* b_ActiveIcon */
+  FvwmPicture *pressicon;  /* b_PressIcon */
+  Window IconWin;          /* b_Swallow */
   Window PanelWin;         /* b_Panel */
   Window BackIconWin;      /* b_Back && b_IconBack */
 
@@ -199,6 +204,7 @@ struct button_info_struct
     unsigned panel_mapped : 1;    /* b_Panel */
     unsigned panel_iconified : 1; /* b_Panel */
     unsigned panel_shaded : 1;    /* b_Panel */
+    unsigned do_swallow_new : 1;
   } newflags;
 #define SLIDE_UP 'u'
 #define SLIDE_DOWN 'd'
@@ -234,6 +240,9 @@ char *mymalloc(int);
 #define mymalloc(a) safemalloc(a)
 #endif
 void handle_xinerama_string(char *args);
+int LoadIconFile(const char *s, FvwmPicture **p, int cset);
+void SetTransparentBackground(button_info *ub,int w,int h);
+void exec_swallow(char *action, button_info *b);
 
 /* ----------------------------- global variables -------------------------- */
 
@@ -241,7 +250,7 @@ extern Display *Dpy;
 extern Window Root;
 extern Window MyWindow;
 extern char *MyName;
-extern button_info *UberButton,*CurrentButton;
+extern button_info *UberButton, *CurrentButton, *ActiveButton;
 extern Bool is_pointer_in_current_button;
 
 extern char *imagePath;
