@@ -1,3 +1,4 @@
+/* -*-c-*- */
 /* This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
  * the Free Software Foundation; either version 2 of the License, or
@@ -107,7 +108,7 @@ static void handle_buttonevent (XEvent *theEvent, WinManager *man)
 
   b = xy_to_button (man, theEvent->xbutton.x, theEvent->xbutton.y);
   if (b && theEvent->xbutton.button >= 1 &&
-      theEvent->xbutton.button <= NUMBER_OF_MOUSE_BUTTONS) {
+      theEvent->xbutton.button <= NUMBER_OF_EXTENDED_MOUSE_BUTTONS) {
     win = b->drawn_state.win;
     if (win != NULL) {
       ConsoleDebug (X11, "Found the window:\n");
@@ -176,10 +177,10 @@ Window find_frame_window (Window win, int *off_x, int *off_y)
   return win;
 }
 
-/***************************************************************************/
-/*  Event handler routines                                                 */
-/*    everything which has its own routine can be processed recursively    */
-/***************************************************************************/
+/*
+ *  Event handler routines
+ *    everything which has its own routine can be processed recursively
+ */
 
 static void reparentnotify_event (WinManager *man, XEvent *ev)
 {
@@ -301,7 +302,7 @@ void xevent_loop (void)
 	      if (globals.transient) {
 		      grab_pointer (man);
 	      }
-	      
+
 	      ex = ey = 10000;
 	      ex2 = ey2 = 0;
       }
@@ -387,7 +388,7 @@ void xevent_loop (void)
 	      if (man->geometry.dir & GROW_FIXED)
 	      {
 		      man->geometry.rows =
-			      theEvent.xconfigure.height / 
+			      theEvent.xconfigure.height /
 			      man->geometry.boxheight;
 		      if (man->geometry.rows < 1)
 		      {
@@ -413,7 +414,7 @@ void xevent_loop (void)
 	      }
 	      else
 	      {
-		      recreate_bg = False; 
+		      recreate_bg = False;
 	      }
       }
 
@@ -825,6 +826,7 @@ void create_manager_window (int man_id)
   int join_style = JoinRound;
   int i;
   WinManager *man;
+  fscreen_scr_t scr;
   ConsoleDebug (X11, "In create_manager_window\n");
 
   man = &globals.managers[man_id];
@@ -891,6 +893,7 @@ void create_manager_window (int man_id)
       &fscr, FSCREEN_XYPOS,
       &man->managed_g.x, &man->managed_g.y,
       &man->managed_g.width, &man->managed_g.height);
+    scr = FSCREEN_XYPOS;
   }
   else
   {
@@ -898,12 +901,15 @@ void create_manager_window (int man_id)
       NULL, FSCREEN_GLOBAL,
       &man->managed_g.x, &man->managed_g.y,
       &man->managed_g.width, &man->managed_g.height);
+    scr = FSCREEN_GLOBAL;
   }
 
   man->theWindow = XCreateWindow(theDisplay, theRoot, sizehints.x, sizehints.y,
 				 man->geometry.width, man->geometry.height,
 				 0, Pdepth, InputOutput, Pvisual, winattrmask,
 				 &winattr);
+  /* hack to prevent mapping on wrong screen with StartsOnScreen */
+  FScreenMangleScreenIntoUSPosHints(scr, &sizehints);
   XSetWMNormalHints(theDisplay, man->theWindow, &sizehints);
   FShapeSelectInput (theDisplay, man->theWindow, FShapeNotifyMask);
   if (globals.transient)
